/******************************************************************
 Example Diffie-Hellman key exchange.
 Copyright 2005 N. Krawetz.  All rights reserved.
 Released as open source.
 No warranty expressed or implied.

 Two parties generate public and private key pairs.
 They uses these to independently derive the same shared secret key.
 Usage:
   ./dh  :: generates a random set of private keys.
   ./dh [private1] [private2] [prime] [root]
     Specify up to 4 parameters for computation.
 ******************************************************************/

#include <stdlib.h>
#include <stdio.h>

typedef long BIG;	/* a big number (don't overflow bits) */

/*****************************************
 PowMod(): The exponent function pow() can overflow bitspace very quickly.
 But since key exchanges use "pow(x,y)%p", the modulus can be used
 to prevent bitspace overflow.
 *****************************************/
BIG	PowMod	(BIG Num, BIG Exp, BIG Mod)
{
  BIG i;
  BIG Result;
  Result=1; /* any number to the 0 power = 1 */
  for(i=0; i<Exp; i++)
    {
    /* Use modulus to throw away any high-order bits that do not
       impact the result.  This prevents bit overflows. */
    Result = (Result * Num) % Mod;
    }
  return(Result);
} /* PowMod() */

/*****************************************************
 SpanTest(): Brute force test for detecting primitive roots.
 The original key exchange algorithm required primitive roots.
 Using a primitive root, it becomes possible for "some" private
 key x and y, all possible shared secrets can be generated.
 Since all shared secrets are possible, there is no chance of
 a shared secret repeating.  And if secrets do not repeat, then
 it increases the difficulty for an attacker to guess the
 private keys.
 But some the DH-exchanges uses "good enough" numbers instead
 of computing primitive roots.
   - This reduces the number of possible shared secret values.
   - This increases the chances of a brute force crack.
 Interestingly, there are many numbers besides primitive roots
 that also span all possible shared secret values.
 This function checks to see if the "primitive root" spans, even
 if it is not a real primitive root.
 (For the prime number 7919, it may look like all odd numbers are
 good candidates because they span.  But not all odd numbers span.
 For example, the numbers 107 and 111 do not span.  Even though 107 is
 a prime number, it is not a primitive root of 7919.)
 Also, the "prime" number does not need to be prime.  For example,
 21 is not prime (3*7), but the root 5 spans it.
 Returns: 0=not span, 1=spans.
 *****************************************************/
int	SpanTest	(BIG Root, BIG Prime)
{
  BIG Num,Val;
  int *Array;
  int rc=1;

  /*****
   Basic property of a good (pseudo-)primitive root value:
   For each Val between 1 and Prime-1, there is a Num such that
     (Root ^ Num) % Prime = Val.
   The brute-force test:
     Compute every (Root ^ Num) % Prime.
     Make sure each Val is seen once.
   *****/
  Array = (int *)calloc((long)Prime,sizeof(int));
  for(Num=0; (Num < Prime) && (rc == 1); Num++)
    {
    Val = PowMod(Num,Root,Prime);
    Array[Val]++;  /* mark this Val as seen */
    /* Since we loop over |Prime| numbers, either every Val is seen
       once, or (at least) one value is seen more than once (and some
       other value is never seen). */
    if (Array[Val] > 1) rc=0; /* was it seen more than once? */
    }
  free(Array);
  return(rc);
} /* SpanTest() */

/*******************************************************************/
int main(int argc, char *argv[])
{
  /* Diffie-Hellman is a key exchange between two parties: a and b */
  BIG a,A;     /* a public/private pair */
  BIG b,B;     /* b public/private pair */
  BIG SSa;     /* shared secret (derived by a) */
  BIG SSb;     /* shared secret (derived by b) */
  BIG Prime;   /* public prime */
  BIG Root;    /* public primitive root, often called a Generator */
  int i;

  /* initialize */
  srand(time(NULL)); /* use rand() if no parameters from command-line */

  Prime = 7919; /* small prime for demonstration only */
  /***************************************
   7919 is chosen for speed and simplicity.
   For strong cryptography, a longer prime number should be
   chosen, such as this 300-digit prime number:
   20395687835640197740576586692903457728019399331434\
   82630947726464532830627227012776329366160631440881\
   73312372882677123879538709400158306567338328279154\
   49969836607190676644003707421711780569087279284814\
   91120222863321448761833763265120835748216479339929\
   61249917319836219304274280243803104015000563790123
   When using any prime number P, there are P*P combinations
   for an attacker.  (7919*7919 is not a serious challenge.)
   The larger the prime number, the bigger the challenge.
   ***************************************/
  if (argc > 3) Prime = atof(argv[3]);
  Root = 3; /* a good primitive root for 7919 */
  if (argc > 4) Root = atof(argv[4]);

  /***************************************
   Root should be a primitive root of Prime.
   Basically that means:
	Root < Prime
	Greatest Common Denominator GCD(Root,Prime) = 1;
	For every value V between 1 and Prime-1, there should be a
	  number N such that (Root^N) mod Prime = V
   Usually, Prime is very large, and Root is very small.
   ***************************************/
  printf("Prime is %ld\n",(long)Prime); 
  printf("Primitive Root is %ld\n",(long)Root);
  if (SpanTest(Root,Prime) != 1)
    {
    printf("WARNING: %ld does not cover all shared secret values.\n",
	(long)Root);
    }

  /******************************************************************
   How big should secrets be?
   The secrets values, a and b, do not need to be larger than the
   Prime.  Larger secrets simply repeat the shared secret sequence.
   The pattern repeats every "Prime-1" times.
   For example, if the Prime is 7919 then (assume all values except a are
   constant), the shared secret should repeat at SS(a) = SS(a+7918):
       Shared secret from a=1 is the same as a=7919
       Shared secret from a=2 is the same as a=7920
       Shared secret from a=3 is the same as a=7921
       Shared secret from a=4 is the same as a=7922
       ...
       Shared secret from a=7918 is the same as a=15836
   Or as a more explicit example:
     let y=15
     let Root=3
     let Prime=7919
     SS(a=1) = SS(a=7919) = SS(a=15837) = 7598
     SS(a=2) = SS(a=7920) = SS(a=15838) = 94
     SS(a=3) = SS(a=7921) = SS(a=15839) = 1502
     etc.
   ******************************************************************/
  if (argc > 1) a = atof(argv[1]);
  else a = rand() % Prime;
  A = PowMod(Root,a,Prime); /* (Root^a)%Prime */
  printf("(a,A) = (%ld,%ld)\n",(long)a,(long)A);

  if (argc > 2) b = atof(argv[2]);
  else b = rand() % Prime;
  B = PowMod(Root,b,Prime); /* (Root^b)%Prime */
  printf("(b,B) = (%ld,%ld)\n",(long)b,(long)B);

  /* exchange public A and B (privates a and b are not shared) */

  /******************************************************************
   What does an attacker know?
   Attacker knows: A, B, Prime, and Root.
   Attacker must derive a from A=(Root^a)%Prime.
   There are ||Prime|| different combinations.
   In this example, there are 7919 different values.
   In real life, the prime may be VERY large -- it just needs to be
   large enough to lead to an unrealistic search.

   What benefits an attacker?
   - Computers are fast.  Finding a large enough prime can be hard.
   - If Root is not a true primitive root, then values duplicate.
     There are many combinations of a and b that lead to the same shared
     secret.  An attacker only needs to find one combination.
   - If Root is not a true primitive root, then values are skipped.
     An attacker does not need to scan all values.
   - If private a or b is too small, then it can be guessed.
   - If private a or b is ZERO then A/B becomes ONE and SS=1.
   - If public A or B == Root then a/b = ONE.
   - If public A and B are exchanged during an algorithm handshake, then
     an attacker can become a man-in-the-middle.  Diffie-Hellman does a
     key exchange but does not provide authentication.
   ******************************************************************/

  SSa = PowMod(B,a,Prime); /* (B^a)%Prime */
  SSb = PowMod(A,b,Prime); /* (A^b)%Prime */
  if (SSa == SSb)
	{
	printf("Shared secret = %ld\n",(long)SSa);
	if (SSa == 0) printf("Shared secret is zero!\n");
	if (SSa == 1) printf("Shared secret is one!\n");
	}
  else
	{
	/* Should never get here! */
	printf("*** Secrets do not match!\n");
	printf("A: Shared secret = %ld\n",(long)SSa);
	printf("B: Shared secret = %ld\n",(long)SSb);
	}

  /******************************************************************
   Now what?
   After exchanging shared secret, the SS can be used to encrypt data
   or initialize stronger privacy algorithms.
   Since SS is only known by a and b, they can use it however they
   want.  Usually it is used with other public information to initialize
   a strong encryption system.  For example, RSA encryption combines the
   shared secret with a public "modulus" value.
   ******************************************************************/
} /* main() */

