/*****************************************************************
 ARP Promiscuous Detection
 Copyright 2005 N. Krawetz  All rights reserved.

 Reference:
   Detecton of Promiscuous Nodes Using ARP Packets, Version 1.0
   by Daiji Sanai, translated by Kelvin King-Pang Tsang
   31-Aug-2001.

 Purpose:
   Send an ARP packet with an invalid hardware address, then
   see if a host responds.  Host in promiscuous mode should respond.

 Usage:
   arp_pd <interface> <target_ip>
     This will send an arp packet for the target_ip.
     If it gets a reply, then the target must be in promiscuous mode.
     The interface specifies which interface to use for accessing the target.

 Things that can stop a reply from being received:
   - The target_ip is not in promiscuous mode.
   - A bridge, switch, or other network device between the source and
     destination is smart enough to see and reject the unknown
     destination hardware address.  (Good news, it means that the
     promiscuous host cannot sniff your traffic.  Better news: this is
     very rare.)

 Things that can cause false-positive identification:
   - Old NICs that do not support multicast.
   - NICs that default to promiscuous mode (e.g., 3Com 3c905).
     These will test positive even when not in promiscuous mode.

 Other notes:
   - There is no unique test for Win2k/NT that does not return a
     positive match for Linux.  (At least, from my own testing.)
   - Promiscuous detection is heavily dependent on both the network
     card and MAC layer.  The accuracy will vary.
   - Not tested against BSD, OS/2, or other operating systems.
     Only tested against multiple Linux versions, Win98, Win2k, Mac OSX.
     Comments for Win9x/ME and Win2k/NT are from the reference material.

 For BSD and similar operating systems (e.g., Mac OS X), refer to
 the Berkeley Packet Filter man page, BPF(4), for writing raw packets.
 *****************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>

struct sockaddr_arp
  {
  short arp_family;
  short arp_protocol;
  int arp_ifindex;
  short arp_hatype;
  char arp_pkttype;
  char arp_halen;
  char arp_addr[8];
  };

struct packet
  {
  struct ether_header ether_header;
  struct arphdr arp_header;
  /* ARP data */
  unsigned char arp_sha[ETHER_ADDR_LEN]; /* Sender hardware address */
  unsigned char arp_sip[4];        /* Sender IP address */
  unsigned char arp_tha[ETHER_ADDR_LEN]; /* Target hardware address */
  unsigned char arp_tip[4];        /* Target IP address */
  /* all above should be 42 bytes; need 60 bytes total */
  unsigned char Trailer[18];       /* Remainder of the packet */
  };
typedef struct packet packet;

/*********************************************************
 Hex2Bin(): Convert a set of ASCII hex codes to binary.
 This skips over all non-hex codes.
 *********************************************************/
void	Hex2Bin	(char *String, char *Binary, int MaxBinaryLen)
{
  int Blen,Slen;	/* indexes into String and Binary */
  int BitsFilled;	/* number of bits filled per byte */
  int Val;	/* value of String[Slen] converted to binary */

  Blen=0;
  Slen=0;
  BitsFilled=0;
  memset(Binary,0,MaxBinaryLen);

  while((Blen < MaxBinaryLen) && (String[Slen] != '\0'))
    {
    if (isxdigit(String[Slen]))
	{
	/* load the character */
	if (isdigit(String[Slen]))	Val = String[Slen] - '0';
	else if  (isupper(String[Slen]))	Val = String[Slen] - 'A' + 10;
	else	Val = String[Slen] - 'a' + 10;
	/* store the character */
	Binary[Blen] = Binary[Blen] * 0x10 + Val;
	BitsFilled += 4;
	if (BitsFilled > 7)
		{
		Blen++;
		BitsFilled=0;
		}
	}
    Slen++;
    }
} /* Hex2Bin() */

/*********************************************************
 LoadInterface(): A computer may have multiple network cards.
 Load the interface for the desired card.
 This will be used for transmitting the poison ARP packet.
 *********************************************************/
void	LoadInterface	(int Socket, char *InterfaceName,
			 struct sockaddr_arp *SocketData,
			 packet *Packet)
{
  struct ifreq Interface;
  struct ifreq *InterfacePointer;	/* pointer to an interface */
  struct ifconf InterfaceConfig;
  char Buffer[0x8000];	/* for interface configuration (make it big) */
  int i;

  memset(SocketData,0,sizeof(struct sockaddr_arp));
  SocketData->arp_protocol = htons(ETHERTYPE_ARP);
  SocketData->arp_halen    = ETHER_ADDR_LEN;

  strcpy(Interface.ifr_name,InterfaceName);

  if (ioctl(Socket,SIOCGIFINDEX,&Interface) < 0)
    {
    printf("Error obtaining network interface %s index\n",Interface.ifr_name);
    exit(-1);
    }
  SocketData->arp_ifindex  = Interface.ifr_ifindex;

  if (ioctl(Socket,SIOCGIFHWADDR,&Interface) < 0)
    {
    printf("Error obtaining %s hardware address\n",Interface.ifr_name);
    exit(-1);
    }
  memcpy(SocketData->arp_addr,Interface.ifr_hwaddr.sa_data,ETHER_ADDR_LEN);
  memcpy(Packet->arp_sha,SocketData->arp_addr,ETHER_ADDR_LEN);
  memcpy(Packet->ether_header.ether_shost,SocketData->arp_addr,ETHER_ADDR_LEN);

  /** SIOCGIFCONF returns a list of IP addresses.
      We need the right one. **/
  InterfaceConfig.ifc_len = sizeof(Buffer);
  InterfaceConfig.ifc_buf = Buffer;
  if (ioctl(Socket,SIOCGIFCONF,&InterfaceConfig) < 0)
    {
    printf("Error obtaining %s IP address\n",Interface.ifr_name);
    exit(-1);
    }
  /** search every interface for the correct name **/
  for(i=0; i*sizeof(struct ifreq) < InterfaceConfig.ifc_len; i++)
    {
    InterfacePointer=&(InterfaceConfig.ifc_req[i]);
    if (!strcmp(InterfacePointer->ifr_name,Interface.ifr_name))
      {
      /** found name!  Save IP address **/
      memcpy(Packet->arp_sip,InterfacePointer->ifr_addr.sa_data+2,4);
      i=InterfaceConfig.ifc_len;  /* break out of loop */
      }
    }
  InterfacePointer = InterfaceConfig.ifc_req;

#if 0
  /* debugging */
  printf("Interface: %s (#%d)\n",InterfaceName,SocketData->arp_ifindex);
  printf("HW address: %02x:%02x:%02x:%02x:%02x:%02x\n",
    Packet->arp_sha[0],Packet->arp_sha[1],Packet->arp_sha[2],
    Packet->arp_sha[3],Packet->arp_sha[4],Packet->arp_sha[5]);
  printf("IP address: %d.%d.%d.%d\n",
    Packet->arp_sip[0],Packet->arp_sip[1],
    Packet->arp_sip[2],Packet->arp_sip[3]);
#endif
} /* LoadInterface() */

/*********************************************************
 DoArp(): Generate an arp packet and wait for a reply.
 Returns 1 if a reply is received, 0 if no reply.
 *********************************************************/
int	DoArp	(int Socket, unsigned char HWcast[ETHER_ADDR_LEN],
		 char *InterfaceName, char *DestinationString)
{
  packet Packet;
  struct sockaddr_arp SocketData;
  in_addr_t IPaddress;
  int rc;
  unsigned char HWBlank[ETHER_ADDR_LEN] = {0x00,0x00,0x00,0x00,0x00,0x00};
  int Buffer[256];
  time_t Timeout;

  /* Clear packet */
  memset(&Packet,0,sizeof(packet));

  /* Get the interface */
  /** This also loads in the source hardware and IP addresses **/
  LoadInterface(Socket,InterfaceName,&SocketData,&Packet);

  /* load Ethernet type and destination */
  Packet.ether_header.ether_type = htons(ETHERTYPE_ARP);
  memcpy(Packet.ether_header.ether_dhost,HWcast,ETHER_ADDR_LEN);

  /* load ARP header (ARP request) */
  Packet.arp_header.ar_hrd = htons(ARPHRD_ETHER);
  Packet.arp_header.ar_pro = htons(ETH_P_IP);
  Packet.arp_header.ar_hln = ETHER_ADDR_LEN;
  Packet.arp_header.ar_pln = 4;
  Packet.arp_header.ar_op  = htons(ARPOP_REQUEST);

  /*** ARP query contents ***/
  /** Set the source ARP and IP to match the sender **/

  /** Set the target ARP and IP address **/
  memcpy(Packet.arp_tha,HWBlank,ETHER_ADDR_LEN);
  IPaddress = ntohl(inet_addr(DestinationString));
  Packet.arp_tip[0] = (IPaddress & 0xff000000) >> 24;
  Packet.arp_tip[1] = (IPaddress & 0x00ff0000) >> 16;
  Packet.arp_tip[2] = (IPaddress & 0x0000ff00) >> 8;
  Packet.arp_tip[3] = (IPaddress & 0x000000ff) >> 0;

  /* Send data out of the socket.  sendto() does not require a connection */
  rc = sendto(Socket,&Packet,sizeof(Packet),0,
  	      (struct sockaddr *)(&SocketData),sizeof(SocketData));
  if (rc != sizeof(packet))
	{
	printf("Write failed -- wrote %d out of %d bytes.\n",rc,sizeof(packet));
	}

  /* Now wait for a reply...  (wait for 5 seconds) */
  Timeout = time(NULL) + 5;
  while(time(NULL) <= Timeout)
    {
    rc = recv(Socket,Buffer,sizeof(packet),0);
    if ((rc >= sizeof(packet))
	  && (((packet *)Buffer)->arp_header.ar_hrd == htons(ARPHRD_ETHER))
	  && (((packet *)Buffer)->arp_header.ar_pro == htons(ETH_P_IP))
	  && (((packet *)Buffer)->arp_header.ar_op == htons(ARPOP_REPLY))
	  )
    	{
	return(1);
	}
    usleep(1000);
    }
  return(0);
} /* DoArp() */

/*********************************************************
 Usage(): Display program usage.
 *********************************************************/
void	Usage	(char *Name)
{
  printf("Usage: %s <interface> <target_ip>\n",Name);
  printf("  This will send an arp packet to the target system.\n");
  printf("  If it gets a reply, then the target must be in promiscuous mode.\n");
  printf("Sample usage:\n");
  printf("  %s eth0 9.9.8.8\n",Name);
  printf("  It will use the network interface 'eth0'.\n");
  printf("  This will send a query for host 9.9.8.8\n");
  printf("  If the host is in promiscuous mode, it will reply.\n");
} /* Usage() */

/*********************************************************/
int	main	(int argc, char *argv[])
{
  int Socket;
  int SocketFlag;

  /* First address: "01" = multicast, "FF" = broadcast */

  /* Check1: seems to work on Linux 2.2/2.4/2.6 only */
  unsigned char HWcastCheck1[ETHER_ADDR_LEN] = {0x01,0x00,0x5E,0x00,0x00,0x00};

  /* Check2: works for all */
  unsigned char HWcastCheck2[ETHER_ADDR_LEN] = {0xFF,0xFF,0x00,0x00,0x00,0x00};

  /* Check3: seems to work on Win9x/ME and Linux 2.2/2.4/2.6, NOT Win2k/NT */
  /* Use this for detecting Win2k */
  unsigned char HWcastCheck3[ETHER_ADDR_LEN] = {0xFF,0x00,0x00,0x00,0x00,0x00};

  /* Check4: works for Windows, but always is positive for Linux */
  unsigned char HWcastCheck4[ETHER_ADDR_LEN] = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFE};
  int rc;

  /* check parameters */
  if (argc != 3) { Usage(argv[0]); exit(-1); }

  /* check user */
  if ((getuid() != 0) && (geteuid() != 0))
    {
    printf("ERROR: This program requires root access.\n");
    exit(-1);
    }

  /* Prepare a socket */
  Socket = socket(PF_PACKET,SOCK_RAW,htons(ETHERTYPE_ARP));
  if (Socket < 0)
    {
    printf("ERROR: Failed to open socket.\n");
    exit(-1);
    }
  /* turn off blocking */
  SocketFlag = fcntl(Socket,F_GETFL);
  if (fcntl(Socket,F_SETFL, SocketFlag | O_NONBLOCK) < 0)
    {
    printf("ERROR: Unable to set non-blocking flag.\n");
    exit(-1);
    }

  /* send two arp requests */
  if (DoArp(Socket,HWcastCheck1,argv[1],argv[2]))
    {
    printf("%s IS (likely) IN PROMISCUOUS MODE!\n",argv[2]);
    printf("Operating system appears to be Linux/BSD.\n");
    }
  else if (DoArp(Socket,HWcastCheck2,argv[1],argv[2]))
    {
    /* ok, could be in promiscuous mode */
    /* let's distinguish Win9x/ME from Linux */
    if (!DoArp(Socket,HWcastCheck3,argv[1],argv[2]))
      {
      /* didn't work, so must be Win2k */
      printf("%s IS IN PROMISCUOUS MODE\n",argv[2]);
      printf("Operating system appears to be Win2k/NT.\n");
      }
    else if (DoArp(Socket,HWcastCheck3,argv[1],argv[2]))
      {
      /* did work, so must be either Win9x/ME or Linux */
      printf("%s could be in promiscuous mode\n",argv[2]);
      printf("If the operating system is Win9x/ME, then it is.\n");
      printf("If the operating system is Linux/BSD then it is not.\n");
      }
    else
      {
      printf("%s IS (likely) IN PROMISCUOUS MODE!\n",argv[2]);
      printf("Operating system appears to be Win9x/ME.\n");
      }
    }
  else
    {
    printf("%s is not detected to be in promiscuous mode.\n",argv[2]);
    }

  close(Socket);
  return(0);
} /* main() */

