/**************************************************************
 Copyright 2005 N. Krawetz
 ICMP gen: Send an ICMP request and wait for any/all replies.
 By default, this performs a ping operation.
 **************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#define	MAXPACKET	4096

struct ICMPTYPE
  {
  int Id;
  char *Name;
  };
struct ICMPTYPE IcmpType[] = {
  {0,"Echo Reply"},
  /* Types 1 and 2 are undefined */
  {3,"Unreachable"},
  {4,"Source Quench"},
  {5,"Redirect"},
  {6,"Alternate Host Address"}, /* RFC1700 */
  /* Type 7 is undefined */
  {8,"Echo Request"},
  {9,"Router Adversitement"},
  {10,"Router Solicitation"},
  {11,"Time Exceeded"},
  {12,"Parameter Problem"},
  {13,"Timestamp Request"},
  {14,"Timestamp Reply"},
  {15,"Information Request"},
  {16,"Information Reply"},
  {17,"Mask Request"},
  {18,"Mask Reply"},
  /* RFC1700 lists values above 18: */
  {30,"Traceroute (RFC1393)"},
  {31,"Datagram Conversion Error (RFC1475)"},
  {32,"Mobile Host Redirect"},
  {33,"IPv6 Where-Are-You"},
  {34,"IPv6 I-Am-Here"},
  {35,"Mobile Registration Request"},
  {36,"Mobile Registration Reply"},
  {-1,"Undefined"}
  };

/**************************************************
 IcmpTypeString(): Find the type of ICMP packet.
 Returns a pointer to a static string.
 **************************************************/
char *	IcmpTypeString	(int Type)
{
  int i;
  for(i=0; IcmpType[i].Id != -1; i++)
    {
    if (Type == IcmpType[i].Id) return(IcmpType[i].Name);
    }
  return(IcmpType[i].Name);
} /* IcmpTypeString() */

/**************************************************
 IcmpCksum(): Compute ICMP checksum.
 Packet includes the ICMP header and data.
 Returns the checksum in network-byte order.
 **************************************************/
int	IcmpCksum	(unsigned char *Packet, int PacketSize)
{
  int i;
  long Sum=0;	/* at least 32 bits */
  int FinalCksum;	/* at least 16 bits */
 
  /* Algorithm: Add all 16-bit words */
  for(i=0; i<PacketSize-1; i+=2)
    {
    if (i==2) continue; /* skip the stored checksum value */
    /* add to the checksum */
    Sum = Sum + ((Packet[i]*256) + Packet[i+1]);
    }
  /* Take care of any odd numbered packets */
  if (PacketSize % 2) { Sum += (Packet[i]*256); }

  Sum = (Sum >> 16) + (Sum & 0xffff); /* add upper word to lower */
  Sum += (Sum >> 16); /* add carry */
  FinalCksum = ((~Sum) & 0xffff);  /* one's compliment */
  return(htons(FinalCksum)); /* return in network-byte order */
} /* IcmpCksum() */

/**************************************************
 Usage(): Display program usage.
 **************************************************/
void	Usage	(char *Name)
{
  int i;
  printf("Usage: %s [options] destination\n",Name);
  printf("  Send an ICMP packet to the destination IP address.\n");
  printf("  Displays all replies.\n");
  printf("    destination :: IP address for sending the request.\n");
  printf("                   Use 255.255.255.255 for a general broadcast.\n");
  printf("  Options:\n");
  printf("    -c num :: set the ICMP checksum to num (for testing invalid checksums)\n");
  printf("    -p num :: wait for up to num packets.  (Default: 1; -1=no limit).\n");
  printf("    -w sec :: wait time in seconds before terminating.  (Default: 3).\n");
  printf("    -S num :: specify ICMP sequence number (default: 0)\n");
  printf("    -P :: display entire packet\n");
  printf("    -T num :: specify ICMP type (default: 8=echo request)\n");
  for(i=0; IcmpType[i].Id != -1; i++)
    {
    printf("        Type %d = %s\n",IcmpType[i].Id,IcmpType[i].Name);
    }
  printf("    -C num :: specify ICMP code (default: 0)\n");
  printf("       See /usr/include/netinet/ip_icmp.h for types and codes\n");
} /* Usage() */

/*********************************************************/
int	main	(int argc, char *argv[])
{
  /* for socket */
  int IcmpSocket;	/* socket for ICMP data */
  struct sockaddr_in SockAddr; /* where to send data */
  int SockAddrLen;
  int SockFlags=0;	/* socket flags */

  /* for options */
  int C;
  int ForceCksum=-1; /* allow forcing the checksum to be a specific value */
  int NumReplies=1; /* number of echo replies to wait for */
  int WaitSec=3; /* number of seconds to wait */
  int TimeEnd;

  /* for packet */
  unsigned char Packet[MAXPACKET]; /* packet header and data */
  int PacketSize;	/* total size of the packet */
  int XferSize;	/* size of the packet transfer (send/receive) */
  struct icmp *IcmpHeader;
  int Type=ICMP_ECHO;	/* ICMP type */
  int Code=0;		/* ICMP code */
  int Sequence=0;	/* ICMP sequence number, for sending many packets */
  int RawPacket=0;	/* display raw packet? */

  /* process command-line */
  while((C=getopt(argc,argv,"c:p:w:T:C:S:P")) != -1)
    {
    switch(C)
	{
	case 'c':
		ForceCksum=htons((~atoi(optarg))&0xffff); /* 1's compliment */
		break;
	case 'p':	NumReplies=atoi(optarg);	break;
	case 'w':	WaitSec=atoi(optarg);	break;
	case 'T':	Type=atoi(optarg);	break;
	case 'C':	Code=atoi(optarg);	break;
	case 'S':	Sequence=atoi(optarg);	break;
	case 'P':	RawPacket=1;	break;
	default:
		printf("Bad option: %c\n",C);
		Usage(argv[0]);
		exit(-1);
	}
    } /* while getopt */
  if (optind != argc-1) { Usage(argv[0]); exit(-1); }

  /*******************/
  /* must run as root */
  if (geteuid() != 0)
  	{
	fprintf(stderr,"ERROR: Must run as root\n");
	exit(-1);
	}

  /*******************/
  /* Prepare the socket */
  IcmpSocket = socket(AF_INET,SOCK_RAW,IPPROTO_ICMP);
  if (IcmpSocket == -1)
  	{
	fprintf(stderr,"ERROR: Unable to open socket\n");
	exit(-1);
	}

  /* Enable support for broadcast packets */
  SockFlags = 1;
  if (setsockopt(IcmpSocket,SOL_SOCKET,SO_BROADCAST,
  	&SockFlags,sizeof(SockFlags)) != 0)
	{
	fprintf(stderr,"ERROR: Unable to set socket options\n");
	exit(-1);
	}

  /* Get the destination address */
  memset(&SockAddr,0,sizeof(SockAddr));
  SockAddr.sin_family = AF_INET;
  SockAddr.sin_addr.s_addr = inet_addr(argv[optind]);
  if ((SockAddr.sin_addr.s_addr == INADDR_NONE) &&
  	strcmp(argv[optind],"255.255.255.255"))
  	{
	fprintf(stderr,"ERROR: invalid IP address \"%s\"\n",argv[optind]);
	exit(-1);
	}
  SockAddr.sin_port = 0; /* unused for ICMP */
  SockAddrLen = sizeof(SockAddr);

  /*******************/
  /* create the packet to transmit: icmp echo request */
  memset(Packet,0,sizeof(Packet));
  IcmpHeader=(struct icmp *)Packet;
  IcmpHeader->icmp_type = Type;
  IcmpHeader->icmp_code = Code;
  IcmpHeader->icmp_id = getpid() & 0xffff; /* process ID */
  /* We use process ID as an identifier so we can distinguish our pings
     from some other process' pings */
  IcmpHeader->icmp_seq = htons(Sequence);

  /* for our echo, don't bother sending data with the request */
  PacketSize = sizeof(struct icmp);
  if (ForceCksum == -1)
       IcmpHeader->icmp_cksum = IcmpCksum(Packet,PacketSize);
  else IcmpHeader->icmp_cksum = ForceCksum;

  /*******************/
  /* send the packet */
  XferSize = sendto(IcmpSocket,Packet,PacketSize,0,
  		    (struct sockaddr *)(&SockAddr),SockAddrLen);
  if (XferSize != PacketSize)
	{
	perror("ERROR");
	fprintf(stderr,
		"ERROR: Unable to send data (sent %d out of %d bytes)\n",
		XferSize,PacketSize);
	exit(-1);
	}

  /*******************/
  /* wait for a reply (or timeout)  */
  TimeEnd=time(NULL)+WaitSec+1; /* since it could be +/- 1 second */
  while((time(NULL) < TimeEnd) && (NumReplies != 0))
    {
    SockAddrLen=sizeof(SockAddr);
    PacketSize=recvfrom(IcmpSocket,Packet,sizeof(Packet),
    			MSG_TRUNC|MSG_DONTWAIT,
  			(struct sockaddr *)(&SockAddr),&SockAddrLen);
    if (PacketSize > 0)
	{
	/* Received packet starts at the length in the IP header */
	PacketSize;
	IcmpHeader=(struct icmp *)(Packet+20);

	/* display the packet */
	printf("%d bytes from %s: ",
		PacketSize+20,inet_ntoa(SockAddr.sin_addr));
	printf("ttl=%d icmp_type=%d (%s) icmp_code=%d\n",
		(int)(Packet[8])&0xff,
		IcmpHeader->icmp_type,IcmpTypeString(IcmpHeader->icmp_type),
		IcmpHeader->icmp_code);

	/* check the packet results */
	if (IcmpCksum(Packet+20,PacketSize) != IcmpHeader->icmp_cksum)
		{
		printf("  Bad checksum: wanted %04x, received %04x\n",
			ntohs(IcmpCksum(Packet+20,PacketSize)),
			ntohs(IcmpHeader->icmp_cksum));
		}

	if (Type == ICMP_ECHO)
	  {
	  if (IcmpHeader->icmp_type != ICMP_ECHOREPLY)
		{
		printf("  Wrong type: received %d (%s), ",
			IcmpHeader->icmp_type,
			IcmpTypeString(IcmpHeader->icmp_type));
		printf("wanted %d (%s)\n",
			ICMP_ECHOREPLY,
			IcmpTypeString(ICMP_ECHOREPLY));
		}
	  }

	if (IcmpHeader->icmp_seq != htons(Sequence))
	  {
	  printf("  Wrong ICMP sequence: received icmp_seq=%d, wanted=%d\n",
		ntohs(IcmpHeader->icmp_seq),Sequence);
	  }
	if (IcmpHeader->icmp_id != (getpid() & 0xffff))
	  {
	  printf("  Wrong ICMP ID: received icmp_id=%d, wanted=%d\n",
	  	IcmpHeader->icmp_id,(getpid() & 0xffff));
	  }
	else if (NumReplies != -1) NumReplies--;

	if (RawPacket)
		{
		printf("  IP:\n   ");
		for(C=0; C < 20; C++) printf(" %02x",Packet[C]);
		printf("\n");
		printf("  ICMP header:\n   ");
		for(C=0; C < 8; C++)
			{
			printf(" %02x",Packet[C+20]);
			if ((C % 16 == 15) && (C+1 < PacketSize))
				printf("\n   ");
			}
		printf("\n");
		if (C < PacketSize)
		  {
		  printf("  ICMP data:\n   ");
		  for(C=0; C < PacketSize-8; C++)
			{
			printf(" %02x",Packet[C+20+8]);
			if ((C % 16 == 15) && (C+1 < PacketSize))
				printf("\n   ");
			}
		  printf("\n");
		  }
		}
	} /* if received packet */
    } /* while receiving */

  return(0);
} /* main() */

