/***************************************************************
 DNS Daemon: Very simple DNS server, based on RFC1035.
 Copyright 2005 N. Krawetz.
 Released as open source.
 No warranty expressed or implied.

 This code is a very simple DNS server.
 It does not implement many of the required elements, such as
 using strings within hostnames (www.hello.com is 3 strings, not one)
 or multiple strings for TXT replies.
 This code is provided as an example only.
 ***************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>

/*** DNS packet header format ***/
struct DNS_RR /* request/reply */
  {
  u_int16_t ID;	/* session serial number */
  u_int8_t Flags;	/* see FLAGs */
  u_int8_t Rcode;	/* see RCODE */
  u_int16_t Qcount;	/* # entries in the question section */
  u_int16_t Acount;	/* # entries in the answer section */
  u_int16_t NScount;	/* # name server records in authority section */
  u_int16_t ARcount;	/* # resource records in additional records section */
  /* NOTE: MTU for UDP is 512 bytes.  512 bytes - header = 500 data bytes */
  unsigned char Data[500];	/* data */
  };
typedef struct DNS_RR DNS_RR;

/*** Flags for DNS header.  OR these together. ***/
#define FLAG_REPLY 0x80	/* is this a query or reply? 0=query, 1=reply */
#define FLAG_OPCODE_MASK 0x30	/* query mask */
#define FLAG_OPCODE_QUERY 0x00	/* standard query */
#define FLAG_OPCODE_IQUERY 0x10	/* inverse query */
#define FLAG_OPCODE_STATUS 0x20	/* server status request */
/* other opcode values bits reserved */
#define FLAG_AA 0x04	/* authoritative answer */
#define FLAG_TC 0x02	/* message truncated */
#define FLAG_RD 0x01	/* recursion denied */

/* Flags added to the rcode byte */
#define FLAG_RA 0x80	/* recursion available */
#define FLAG_AAA 0x20	/* answer authenticated */
#define RCODE_MASK 0x0f
enum RCODE
  {
  RCODE_NO_ERROR=0,	/* no error condition */
  RCODE_FORMAT_ERROR,	/* format error */
  RCODE_SERVER_ERROR,	/* server error */
  RCODE_NAME_ERROR,	/* name error */
  RCODE_NA,	/* not implemented (not available) */
  RCODE_REFUSED,	/* refused */
  };
typedef enum RCODE RCODE;

enum TYPE
  {
  TYPE_A=1,	/* host address */
  TYPE_NS,	/* authoritative name server */
  TYPE_MD,	/* mail destination (Obsolete - use MX) */
  TYPE_MF,	/* mail forwarder (Obsolete - use MX) */
  TYPE_CNAME,	/* canonical name for an alias */
  TYPE_SOA,	/* start of a zone of authority */
  TYPE_MB,	/* mailbox domain name (EXPERIMENTAL) */
  TYPE_MG,	/* mail group member (EXPERIMENTAL) */
  TYPE_MR,	/* mail rename domain name (EXPERIMENTAL) */
  TYPE_NULL,	/* a NULL RR (EXPERIMENTAL) */
  TYPE_WKS,	/* well known service description */
  TYPE_PTR,	/* domain name pointer */
  TYPE_HINFO,	/* host information */
  TYPE_MINFO,	/* mailbox or mail list information */
  TYPE_MX,	/* mail exchange */
  TYPE_TXT,	/* text strings */
  TYPE_AAAA=0x1c,	/* IPv6 address request - RFC1886 */
  /** QTYPE are a superset of TYPE **/
  QTYPE_AXFR=252,	/* request for a transfer of an entire zone */
  QTYPE_MAILB=253,	/* request for mailbox-related records (MB/MG/MR) */
  QTYPE_MAILA=254,	/* request for mail agent RRs (Obsolete - see MX) */
  QTYPE_ALL=255 	/* request for all records (symbol "*") */
  };
typedef enum TYPE TYPE;

enum CLASS
  {
  CLASS_IN=1,	/* Internet class */
  CLASS_CS,	/* CSNET class (Obsolete) */
  CLASS_CH,	/* CHAOS class */
  CLASS_HS,	/* Hesiod class */
  /** QCLASS are a superset of CLASS **/
  QCLASS_ALL=255	/* request for any class (symbol "*") */
  };
typedef enum CLASS CLASS;

/**********************************************************
 DisplayPacket(): Display the contents of a DNS packet.
 **********************************************************/
void	DisplayPacket	(DNS_RR Packet, int PacketLen)
{
  int Qcount,Acount,NScount,ARcount;
  int ProcessLen;
  int i,j,Len;
  u_int16_t Val;

  if (PacketLen < 12)
    {
    printf("Invalid DNS packet.\n");
    return;
    }
  /* display header */
  printf("DNS sequence identifier: %d\n",ntohs(Packet.ID));
  printf("  Flags:\n");
  printf("    %s\n",Packet.Flags & FLAG_REPLY ? "Reply" : "Request");
  if (Packet.Flags & FLAG_OPCODE_MASK == FLAG_OPCODE_IQUERY)
	printf("    Inverse query\n");
  else if (Packet.Flags & FLAG_OPCODE_MASK == FLAG_OPCODE_STATUS)
	printf("    Status query\n");
  else	printf("    Standard query\n");
  if (Packet.Flags & FLAG_AA) printf("    Authoritative answer\n");
  if (Packet.Flags & FLAG_TC) printf("    Message truncated\n");
  if (Packet.Flags & FLAG_RD) printf("    Recursion denied\n");

  if (Packet.Rcode & FLAG_RA) printf("    Recursion available\n");
  printf("  Error code: ");
  switch(Packet.Rcode & 0x7f)
    {
    case RCODE_NO_ERROR: printf("no error\n"); break;
    case RCODE_FORMAT_ERROR: printf("format error\n"); break;
    case RCODE_SERVER_ERROR: printf("server error\n"); break;
    case RCODE_NAME_ERROR: printf("name error\n"); break;
    case RCODE_REFUSED: printf("refused\n"); break;
    default: printf("error %d not implmented\n",Packet.Rcode & 0x7f);
    }

  Qcount = ntohs(Packet.Qcount);
  Acount = ntohs(Packet.Acount);
  NScount = ntohs(Packet.NScount);
  ARcount = ntohs(Packet.ARcount);

  /* compute data size */
  PacketLen = PacketLen - 12;
  ProcessLen=0;

  printf("  Number of queries: %d\n",Qcount);
  for(i=0; i<Qcount; i++)
    {
    printf("    Query #%d: ",i+1);
    Len=Packet.Data[ProcessLen];
    ProcessLen++;
    while(Len > 0)
      {
      if (ProcessLen + Len + 4 > PacketLen)
        {
        printf("  Packet overflow.\n");
        return;
        }
      for(j=0; j<Len; j++) fputc(Packet.Data[ProcessLen++],stdout);
      fputc('.',stdout);
      Len=Packet.Data[ProcessLen];
      ProcessLen++;
      }
    Val = Packet.Data[ProcessLen]*256 + Packet.Data[ProcessLen+1];
    printf("  type=%d",Val);
    Val = Packet.Data[ProcessLen+2]*256 + Packet.Data[ProcessLen+3];
    printf("  class=%d\n",Val);
    ProcessLen += 4;
    }

  printf("  Number of answers: %d\n",Acount);
  printf("  Number of server records: %d\n",NScount);
  printf("  Number of additional records: %d\n",ARcount);

  printf("\n");
} /* DisplayPacket() */
 
/**********************************************************/
int	main	(int argc, char *argv[])
{
  int UDPsocket;	/* server socket */
  DNS_RR Request,Reply;
  int RequestLen,ReplyLen;
  struct sockaddr_in Server;
  struct sockaddr Client;
  int ClientLen;
  int Len;
  int Ipaddr;

  if (argc != 4)
    {
    printf("Simple DNS server that always returns the same reply.\n");
    printf("Usage: %s ipaddress hostname text\n",argv[0]);
    printf("  ipaddress: IP address to reply (1.2.3.4)\n");
    printf("  hostname:  hostname to reply (cow.farm.lan)\n");
    printf("  text:      text comment to reply (\"Text in quotes\")\n");
    exit(-1);
    }

  UDPsocket = socket(AF_INET,SOCK_DGRAM,0);
  if (UDPsocket < 0)
    {
    fprintf(stderr,"ERROR: Unable to open socket.\n");
    exit(-1);
    }

  Server.sin_family = AF_INET;
  Server.sin_addr.s_addr = INADDR_ANY;
  Server.sin_port = htons(53);	/* DNS normally runs on port 53/UDP */
  if (bind(UDPsocket,(struct sockaddr *)(&Server),sizeof(Server)) != 0)
    {
    fprintf(stderr,"ERROR: Failed to bind to port 53/UDP.\n");
    exit(-1);
    }

  /* Now listen for requests and reply */
  while(1)
    {
    ClientLen=sizeof(Client);
    RequestLen = recvfrom(UDPsocket,&Request,sizeof(Request),0,
    			  &Client,&ClientLen);
    if (RequestLen >= 12)	/* 12 bytes is minimum */
      {
      /* Process packet request */
      DisplayPacket(Request,RequestLen);
      memset(&Reply,0,sizeof(Reply));
      Reply.ID = Request.ID;
      Reply.Flags = FLAG_REPLY | FLAG_AA; /* authoritative reply */
      Reply.Rcode = RCODE_SERVER_ERROR; /* assume error */
      ReplyLen = 12;

      /* check if it is a request we can handle */
      if ((Request.Flags & FLAG_REPLY) == 0)
        {
	Reply.Rcode = RCODE_NO_ERROR;
	Reply.Acount = htons(2); /* 2 replies: hostname and TXT */
	Len=0;

	/* Reply #1: hostname and IP address */
	/* store hostname */
	Reply.Data[Len++]=strlen(argv[2]);
	memcpy(Reply.Data+Len,argv[2],strlen(argv[2]));
	Len+=strlen(argv[2]);
	Reply.Data[Len++] = 0; /* end of string */
	Reply.Data[Len++] = (TYPE_A/256)&0xff; /* upper byte of TYPE */
	Reply.Data[Len++] = (TYPE_A)&0xff;     /* lower byte of TYPE */
	Reply.Data[Len++] = (CLASS_IN/256)&0xff; /* upper byte of CLASS */
	Reply.Data[Len++] = (CLASS_IN)&0xff;     /* lower byte of CLASS */
	Reply.Data[Len++] = 0; /* TTL = 4 bytes */
	Reply.Data[Len++] = 0; /* TTL */
	Reply.Data[Len++] = 0; /* TTL */
	Reply.Data[Len++] = 10; /* lowest part of TTL = 10 seconds */
	Reply.Data[Len++] = 0; /* upper part of data length */
	Reply.Data[Len++] = 4; /* lower part of data length = 4 octets */
	Ipaddr = ntohl(inet_addr(argv[1]));
	Reply.Data[Len++] = (Ipaddr >> 24) & 0xff;
	Reply.Data[Len++] = (Ipaddr >> 16) & 0xff;
	Reply.Data[Len++] = (Ipaddr >> 8) & 0xff;
	Reply.Data[Len++] = (Ipaddr) & 0xff;

	/* Reply #2: hostname and TXT string */
	Reply.Data[Len++]=0xc0;	/* make this a pointer */
	Reply.Data[Len++]=0x0c;	/* point to name (start of data) */
	Reply.Data[Len++] = (TYPE_TXT/256)&0xff; /* upper byte of TYPE */
	Reply.Data[Len++] = (TYPE_TXT)&0xff;     /* lower byte of TYPE */
	Reply.Data[Len++] = (CLASS_IN/256)&0xff; /* upper byte of CLASS */
	Reply.Data[Len++] = (CLASS_IN)&0xff;     /* lower byte of CLASS */
	Reply.Data[Len++] = 0; /* TTL = 4 bytes */
	Reply.Data[Len++] = 0; /* TTL */
	Reply.Data[Len++] = 0; /* TTL */
	Reply.Data[Len++] = 10; /* lowest part of TTL = 10 seconds */
	Reply.Data[Len++] = 0; /* upper part of data length */
	Reply.Data[Len++] = strlen(argv[3])+1; /* data length */
	Reply.Data[Len++] = strlen(argv[3]); /* string length */
	memcpy(Reply.Data+Len,argv[3],strlen(argv[3]));
	Len += strlen(argv[3]);

	ReplyLen += Len; /* total size of packet */
	} /* if request */

      DisplayPacket(Reply,ReplyLen);
      if (sendto(UDPsocket,&Reply,ReplyLen,0,&Client,ClientLen) < 0)
	{
	perror("ERROR: sendto failed");
	}
      } /* if at least 12 bytes */
    } /* while(1) */

  /* Should never get here! */
  return(0);
} /* main() */

